@@ -1,5 +1,5 @@
 #
-from .util import timer, hub_authenticated
+from .util import timer, hub_authenticated, HubException
 from dateutil.parser import parse
 from datetime import datetime, timedelta
 from requests.structures import CaseInsensitiveDict
@@ -24,8 +24,8 @@ from . import __version__
 
 from .queries import *
 from .access import save_jwt, delete_jwt, restore_previous_jwt, get_previous_jwt
-from http.server import HTTPServer, BaseHTTPRequestHandler
 from urllib.parse import urlparse, parse_qs
+from urllib3.exceptions import HTTPError
 import time
 
 # import urllib3
@@ -97,10 +97,6 @@ def assets_and_metadata(gqlh, db_index, current_db):
     return assets, metadata, dv_column_key
 
 
-class GraphQLError409(Exception):
-    pass
-
-
 class GraphQLException(Exception):
     pass
 
@@ -137,7 +133,7 @@ class HubClient:
         debug: bool = False,
     ):
         self.__authenticated = False
-        self.__jwt = {}
+        self.__jwt = {"access_token": "none"}
         self.__session_id = str(uuid.uuid4())
         self.__graphql_transport = None
         self.__graphql_client = None
@@ -146,7 +142,7 @@ class HubClient:
         self.__debug = debug
         data_file = hub_data if os.path.isfile(hub_data) else default_hub_data
         self.__default_data = data_file == default_hub_data
-        if debug and (self.__default_data):
+        if debug and self.__default_data:
             print(f"@ Hub data file: {data_file}")
         self.__gqlh, self.__current_db, self.__db_index = initialize_hub_data(data_file)
         self.__current_db_index = 0
@@ -161,6 +157,10 @@ class HubClient:
     def session_id(self) -> str:
         return self.__session_id
 
+    @typechecked
+    def get_jwt(self) -> dict:
+        return self.__jwt
+
     @typechecked
     def set_jwt(self, jwt: dict, gw_url):
         self.__jwt = jwt.copy()
@@ -179,7 +179,7 @@ class HubClient:
     def _id_token(self) -> str:
         token = self.__jwt.get("id_token", None)
         if token is None:
-            raise GraphQLException("@@@ Please (re)start Hub login sequence")
+            raise HubException("@@@ Please (re)start Hub login sequence")
         return token
 
     @typechecked
@@ -201,10 +201,10 @@ class HubClient:
     @typechecked
     def asset_metadata(self, asset: str):
         if asset.lower() not in self.__assets:
-            print(
+            raise HubException(
                 f"@@ error: asset {asset} not in dataset asset list, check hub.assets()"
             )
-            return
+
         return self.__assets_metadata[asset]
 
     def all_assets_metadata(self):
@@ -236,13 +236,12 @@ class HubClient:
         return f"{version} (status: {status})"
 
     @hub_authenticated
-    @typechecked
     def set_dataset(self, dataset: str):
-        try:
-            hub_db_namespaces[dataset]
-        except KeyError:
-            print(f"@@ Dataset {dataset} does not exist, please check hub.datasets()")
-            return
+        if not isinstance(dataset, str):
+            raise HubException(f"@@ Dataset most be a string, please check hub.datasets()")
+        if hub_db_namespaces.get(dataset, None) is None:
+            raise HubException(f"@@ Dataset {dataset} does not exist, please check hub.datasets()")
+
         for j in range(len(self.__gqlh["Database"])):
             if self.__gqlh["Database"][j]["name"] == dataset:
                 self.__current_db_index = j
@@ -255,12 +254,13 @@ class HubClient:
                 break
 
     @hub_authenticated
-    @typechecked
     def namespace_of(self, dataset: str):
-        try:
-            return hub_db_namespaces[dataset]
-        except KeyError:
-            print(f"@@ Dataset {dataset} does not exist, please check hub.datasets()")
+        if not isinstance(dataset, str):
+            raise HubException(f"@@ Dataset most be a string, please check hub.datasets()")
+        if hub_db_namespaces.get(dataset, None) is None:
+            raise HubException(f"@@ Dataset {dataset} does not exist, please check hub.datasets()")
+        return hub_db_namespaces[dataset]
+
 
     @hub_authenticated
     @typechecked
@@ -285,10 +285,10 @@ class HubClient:
     ) -> Union[None, List[str]]:
         if len(asset) > 0:
             if asset.lower() not in self.__assets:
-                print(
+                raise HubException(
                     f"@@ error: asset {asset} not in dataset asset list, check hub.assets()"
                 )
-                return
+
         if not multiple_asset:
 
             def len_test(li):
@@ -356,6 +356,8 @@ class HubClient:
             q_resolved,
             {"id": dataview_id, "namespace": namespace_id, "queryId": "Asset_value"},
         )
+        if len(data_items["dataview"]) == 0:
+            raise HubException(f"@@ Bad namespace ({namespace_id}) and/or dataview ID ({dataview_id})")
         v2_column_key = self.__dv_column_key.get(dataview_id, None)
         column_key = (
             "column_name" if v2_column_key is None else f"{v2_column_key}|column"
@@ -413,7 +415,7 @@ class HubClient:
             ),
         )
         if len(reply["dataview"]) == 0:
-            print("@@ NO DATA: Check dataview_id and/or date range")
+            print("@@ NO DATA: Check namespace_id, dataview_id and/or date range")
             return None, [], None
         result = reply["dataview"][0]["data"]
         return result["nextPage"], result["data"], result["firstPage"]
@@ -465,17 +467,20 @@ class HubClient:
         verbose: bool = False,
         stored: bool = False,
     ):
-        return self.dataview_get_data_pd(
-            namespace_id,
-            dataview_id,
-            start_index,
-            end_index,
-            interval,
-            count,
-            sub_second_interval,
-            verbose,
-            stored,
-        )
+        try:
+            return self.dataview_get_data_pd(
+                namespace_id,
+                dataview_id,
+                start_index,
+                end_index,
+                interval,
+                count,
+                sub_second_interval,
+                verbose,
+                stored,
+            )
+        except GraphQLException as e:
+            raise e
 
     def dataview_get_data_pd(
         self,
@@ -552,17 +557,30 @@ class HubClient:
                     self.__dataview_next_page = None
                     break
                 print("+", end="", flush=True)
+            except HTTPError as e:
+                if "502" in str(e):
+                    print("@", end="")
+                    continue
             except Exception as e:
                 if not any(
-                    ss in str(e) for ss in ["408:", "503:", "504:", "409:", "502:"]
+                    ss in str(e).lower()
+                    for ss in ["408", "503", "504", "409", "502", "unauthenticated"]
                 ):
                     raise e
-                if "409:" in str(e):
+                if "unauthenticated" in str(e).lower():
+                    raise GraphQLException(
+                        "@@@ Please (re)start Hub login sequence (cell with hub_login() )"
+                    )
+                if "409" in str(e):
                     print("#", end="")
                     continue
-                if "408:" not in str(e):
-                    print(f"[restart-{str(e)[3:6]}]", end="")
-                    raise Exception("408")
+                if "502" in str(e):
+                    print("[@]", end="")
+                    continue
+                if "408" not in str(e):
+                    print(f"[restart-{str(e)}]", end="")
+                    raise GraphQLException(f"Got: {str(e)}")
+
                 df = pd.DataFrame()
                 next_page = None
                 if count is None:
@@ -601,6 +619,8 @@ class HubClient:
                 resume=resume,
                 max_stored_rows=max_rows,
             )
+        except GraphQLException as e:
+            raise e
         except Exception as e:
             if "404" in str(e):
                 print(
@@ -638,57 +658,74 @@ class HubClient:
         return self.__graphql_client.execute(query, variable_values=variable_values)
 
 
-jwt = {"access_token": "none"}
+def set_token_and_check(hub, jwt, custom_url):
+    hub.set_jwt(jwt, custom_url)
+    reply = hub.graphql_query(q_endpoint_check)
+    if reply.get("databases", False):
+        hub.set_authenticated()
+        return True
+    else:
+        return False
 
 
-def hub_login(force_login=False, gw_url=None):
-    if force_login:
+def hub_connect(jwt: dict, gw_url: str = None):
+    hub = HubClient()
+    if set_token_and_check(hub, jwt, gw_url):
+        return hub
+    else:
+        raise GraphQLException("@@ Got bad JWT")
+
+
+def login_state(hub, jwt: dict, found_jwt: bool = False):
+    status = "+" if found_jwt else ""
+    status += "@" if not hub.default_data() else ""
+    status += "#" if jwt.get("creds", None) else ""
+    return status
+
+
+def hub_login(force: bool = False, gw_url: str = None):
+    if force:
         delete_jwt()
     hub = HubClient()
     new_tab = 'target="_blank"'
     registration_link = f'(<a {new_tab} href="{REGISTRATION_URL}"><font color="blue">register here</font></a>)'
-    step1 = '<font color="orange"><b>Step 1. Click here to initiate login sequence on new tab</b></font>'
-    login_md = f"""<b>Academic Hub Login {registration_link}, version {__version__}</b> 
+    step1 = '<font color="purple"><b>Click here to initiate login sequence on new tab</b></font>'
+    login_md = f"""![AVEVA banner](https://academichub.blob.core.windows.net/images/aveva-banner.png)<br>
+    <b>Academic Hub Login {registration_link}, version {__version__}</b> 
 
-Follow the 5 steps below:
+Follow the steps below:
 
-<ol start="0">
-  <li><b>If Login status is "OK", your previous login still works, skip Steps 1 to 4</b></li>
-  <li>Click on the Step 1 link below will open a new browser tab</li>
-  <li>Enter "academic-hub" when asked for the organization</li>
-  <li>Select your Google account and enter your credentials</li>
-  <li>Upon successful login with Google, come back here and press the "Click me" button (Step 4)</li>
+<ol>
+  <li><b>If Login status is "OK" (at bottom of this cell), you can continue with the rest of the notebook</b></li>
+  <li>Click on the purple link to open a new browser tab and:</li>
+  <ol type="a">
+     <li>Enter "academic-hub" when asked for the organization</li>
+     <li>Select your Google account and enter your credentials</li>
+     <li>When successfully logged in with AVEVA authentication web page, come back to this notebook</li>
+     <li> <a {new_tab} href="{AUTH_ENDPOINT}/login?hub-id={hub.session_id()}">{step1}</a> </li>
+  </ol>
+  <li>Complete authentication procedure by the clicking button below</li>
 </ol>
-
-&nbsp;&nbsp;&nbsp;&nbsp;<a {new_tab} href="{AUTH_ENDPOINT}/login?hub-id={hub.session_id()}">{step1}</a>
 """
     html = markdown.markdown(login_md)
     button = widgets.Button(
-        description="Step 4. Click me",
+        description="Click me",
         disabled=False,
         button_style="warning",
         tooltip="Click me when Google login is successful",
         icon="cloud-download",
     )
 
-    def set_token_and_check(jwt, custom_url):
-        hub.set_jwt(jwt, custom_url)
-        reply = hub.graphql_query(q_endpoint_check)
-        if reply.get("databases", False):
-            hub.set_authenticated()
-            return True
-        else:
-            return False
-
     restore_previous_jwt(hub.session_id())
     time.sleep(1)
     jwt = get_previous_jwt(hub.session_id())
-    login_status = "--undefined--"
+    login_status = "-- not logged in --"
     try:
-        if set_token_and_check(jwt, gw_url):
-            default_data_indicator = "+" if not hub.default_data() else ""
-            login_status = f"OK, you can proceed+{default_data_indicator}"
-    except GraphQLException:
+        if set_token_and_check(hub, jwt, gw_url):
+            default_data_indicator = "!" if not hub.default_data() else ""
+            default_data_indicator += "@" if jwt.get("creds", None) else ""
+            login_status = f"OK, you can proceed {login_state(hub, jwt, True)}"
+    except HubException:
         pass
     except Exception as e:
         if "unauthenticated" in str(e).lower():
@@ -704,16 +741,14 @@ Follow the 5 steps below:
     )
 
     def button_confirm(_):
-        global jwt
         r = requests.get(
             f"{AUTH_ENDPOINT}/token?jwt=1",
             headers={"hub-id": hub.session_id()},
             verify=False,
         )
         if 200 == r.status_code:
-            if set_token_and_check(eval(r.text), gw_url):
-                jwt = eval(r.text)
-                status.value = "OK, you can proceed"
+            if set_token_and_check(hub, eval(r.text), gw_url):
+                status.value = f"OK, you can proceed {login_state(hub, eval(r.text))}"
             else:
                 status.value = "ERROR: Hub endpoint failed, go to Step 1"
         else:
@@ -723,12 +758,9 @@ Follow the 5 steps below:
     output = widgets.Output()
 
     def on_value_change(change):
-        global jwt
         with output:
             if "OK" in change["new"]:
-                save_jwt(jwt)
-            else:
-                display("oops: bad jwt")
+                save_jwt(hub.get_jwt())
 
     status.observe(on_value_change, names="value")
 
